#
# Revelation 0.2.1 - a password manager for GNOME 2
# http://oss.wired-networks.net/revelation/
#
# Module for importing from / exporting to a datafile
#
#
# Copyright (c) 2003-2004 Erik Grinaker
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#

import gobject, os, revelation, gtk, libxml2, zlib, random, math, time, binascii, re
from Crypto.Cipher import AES, Blowfish
from Crypto.Hash import MD5

TYPE_AUTO		= "autodetect"
TYPE_FPM		= "fpm"
TYPE_REVELATION		= "revelation"
TYPE_XML		= "xml"

OPTION_PASSWORD		= "password"

# data classes
class FileTypes(object):

	def __init__(self):

		self.filetypes = {

			TYPE_FPM	: {
				"name"		: "Figaro's Password Manager",
				"import"	: gtk.TRUE,
				"export"	: gtk.TRUE,
				"datahandler"	: FPMData,
				"encryption"	: gtk.TRUE,
				"checkbuffer"	: 1024,

				"options"	: {
					OPTION_PASSWORD	: ""
				}
			},

			TYPE_REVELATION	: {
				"name"		: "Revelation",
				"import"	: gtk.TRUE,
				"export"	: gtk.FALSE,
				"datahandler"	: RevelationData,
				"encryption"	: gtk.TRUE,
				"checkbuffer"	: 1024,

				"options"	: {
					OPTION_PASSWORD	: ""
				}
			},

			TYPE_XML : {
				"name"		: "XML (eXtensible Markup Language)",
				"import"	: gtk.TRUE,
				"export"	: gtk.TRUE,
				"datahandler"	: RevelationXMLData,
				"encryption"	: gtk.FALSE,
				"checkbuffer"	: 1024,
				
				"options"	: {
				}
			}
		}


	def get_export_types(self):
		"""Returns a list of export types"""

		types = []
		for type, data in self.filetypes.items():
			if data["export"] == gtk.TRUE:
				types.append(type)

		types.sort()
		return types


	def get_import_types(self):
		"""Returns a list of import types"""

		types = []
		for type, data in self.filetypes.items():
			if data["import"] == gtk.TRUE:
				types.append(type)

		types.sort()
		return types


	def get_data(self, type, attr = None):
		"""Returns data for a type"""

		if attr == None:
			return self.filetypes[type]
		else:
			return self.filetypes[type][attr]


	def get_largest_checkbuffer(self, types):
		"""Returns the largest checkbuffer for a list of types"""

		bufsize = 0
		for type in types:
			typebuffer = self.get_data(type, "checkbuffer")
			if typebuffer > bufsize:
				bufsize = typebuffer

		return bufsize


	def get_option_template(self, type):
		"""Returns an option template for a type"""

		if type == None:
			return {}
		else:
			return self.filetypes[type]["options"]



# a few exceptions
class FormatError(Exception):
	"""Exception for invalid file formats"""
	pass

class PasswordError(Exception):
	"""Exception for wrong password"""
	pass

class VersionError(Exception):
	"""Exception for unknown versions"""
	pass

class DetectError(Exception):
	"""Exception for failed autodetection"""
	pass

class EntryError(Exception):
	"""Base class for entry errors"""
	pass

class EntryFieldError(EntryError):
	"""Exception for invalid field type for an entry type"""
	pass

class EntryTypeError(EntryError):
	"""Exception for unknown entry type"""
	pass

class OptionError(Exception):
	"""Exception for invalid / missing options etc"""
	pass


# general classes

class DataFile(gobject.GObject):
	"""Class for handling data files"""

	def __init__(self, file = None, type = None):
		gobject.GObject.__init__(self)
		self.filetypes = FileTypes()
		self.options = {}

		self.set_file(file)
		self.set_type(type)


	# private methods
	def __read(self, file, bytes = -1):
		"""Reads data from a file"""

		if file == None:
			raise IOError

		fp = open(file, "rb", 0)
		data = fp.read(bytes)
		fp.close()

		return data


	def __write(self, file, data):
		"""Writes data to a file"""

		if file == None:
			raise IOError

		fp = open(file, "wb", 0)
		fp.write(data)
		fp.flush()
		fp.close()


	# a few small state methods
	def get_file(self):
		return self.file

	def get_option(self, name):
		if not self.has_option(name):
			raise OptionError

		return self.options[name]

	def get_type(self):
		return self.type

	def has_option(self, name):
		return self.options.has_key(name)

	def set_file(self, file):
		self.file = file

	def set_option(self, name, value):
		if not self.options.has_key(name):
			raise OptionError

		self.options[name] = value

	def set_type(self, type):
		self.type = type

		# set up default options for the type
		oldoptions = self.options
		self.options = self.filetypes.get_option_template(type)

		for key, val in self.options.items():
			if oldoptions.has_key(key):
				self.options[key] = oldoptions[key]


	# main methods
	def check_file(self):
		"""Checks if a file is valid"""

		# check if the file exists and is readable
		if self.file == None or os.access(self.file, os.R_OK) == 0:
			raise IOError


	def check_format(self):
		"""Checks the format of a file"""

		handler = self.filetypes.get_data(self.type, "datahandler")()
		data = self.__read(self.file, self.filetypes.get_data(self.type, "checkbuffer"))
		handler.check_data(data)


	def detect_type(self):
		"""Detects a files type"""

		# get types to check, and load the header
		detecttypes = self.filetypes.get_import_types()
		bufsize = self.filetypes.get_largest_checkbuffer(detecttypes)
		header = self.__read(self.file, bufsize)

		# attempt to detect the file type
		for detecttype in detecttypes:
			handler = self.filetypes.get_data(detecttype, "datahandler")()

			if handler.detect_type(header) == gtk.TRUE:
				self.set_type(detecttype)
				return detecttype

		raise DetectError


	def load(self):
		"""Loads a file, returns an entrystore with its contents"""

		# load file
		self.check_file()
		self.check_format()
		data = self.__read(self.file)

		# import data into entrystore
		entrystore = revelation.data.EntryStore(revelation.data.EntryTypes())
		importer = self.filetypes.get_data(self.type, "datahandler")()
		importer.import_data(entrystore, data, self.options)

		return entrystore


	def save(self, entrystore):
		"""Saves data from an entrystore to the file"""

		exporter = self.filetypes.get_data(self.type, "datahandler")()
		data = exporter.export_data(entrystore, self.options)
		self.__write(self.file, data)



class Data(gobject.GObject):
	"""Base class for data handlers"""

	def __init__(self):
		gobject.GObject.__init__(self)


	def cipher_decrypt(self, data):
		"""Decrypts data"""

		if len(data) % self.cipher.block_size != 0:
			raise FormatError

		return self.cipher.decrypt(data)


	def cipher_encrypt(self, data):
		"""Encrypts data"""

		if len(data) % self.cipher.block_size != 0:
			raise FormatError

		return self.cipher.encrypt(data)


	def cipher_init(self, engine, password, iv = None, blocksize = None, keysize = None, keypad = chr(0)):
		"""Initializes a cipher object"""

		if blocksize != None:
			engine.block_size = blocksize

		if keysize != None:
			engine.key_size = keysize
			padlen = keysize - (len(password) % keysize)
			password = password + (keypad * padlen)

		if iv == None:
			self.cipher = engine.new(password)
		else:
			self.cipher = engine.new(password, engine.MODE_CBC, iv)


	def xml_import_attrs(self, node):
		"""Returns a dict of attributes for a node"""

		attrs = {}

		attr = node.properties
		while attr is not None:
			attrs[attr.name] = attr.content
			attr = attr.next

		return attrs


	def xml_import_init(self, xml):
		"""Initiates an xml document parser"""

		try:
			doc = libxml2.parseDoc(xml)
		except (libxml2.parserError, TypeError):
			raise FormatError

		return doc


	def xml_import_scan(self, node, childname):
		"""Scans a node for a specific child, returns the child if found"""

		child = node.children
		while child is not None and child.name != childname:
			child = child.next

		if child is not None and child.name == childname:
			return child
		else:
			return None



class FPMData(Data):
	"""Class for handling data from Figaros Password Manager"""

	def __init__(self):
		Data.__init__(self)
		self.blocksize = 8


	def __decrypt_field(self, data):
		"""Decodes and decrypts a data field"""

		data = self.__hex_to_bin(data)
		data = self.cipher_decrypt(data)
		data = self.__unrotate_field(data)

		return data


	def __encrypt_field(self, data):
		"""Encrypts and encodes a data field"""

		blocks = (len(data) / (self.blocksize - 1)) + 1
		size = self.blocksize * blocks

		# add noise
		data = data + "\x00"
		data = data + self.__generate_noise(size - len(data))

		# rotate and encrypt field
		data = self.__rotate_field(data)
		data = self.cipher_encrypt(data)

		# ascii-armor data
		data = self.__bin_to_hex(data)

		return data


	def __generate_noise(self, bytes):
		"""Generates random junk, for scrambling data"""

		noise = ""
		for i in range(bytes):
			noise = noise + chr(int(random.random() * 255))

		return noise


	def __generate_salt(self):
		"""Generates a salt for a new password file"""

		salt = ""
		for i in range(4):
			salt = salt + chr(int(random.random() * 255))

		salt = self.__bin_to_hex(salt)
		return salt


	def __bin_to_hex(self, data):
		"""Converts binary data to fpm ascii armored data"""

		hex = ""
		for i in range(len(data)):
			high = ord(data[i]) / 16
			low = ord(data[i]) - high * 16
			hex = hex + chr(ord('a') + high) + chr(ord('a') + low)

		return hex


	def __hex_to_bin(self, data):
		"""Converts fpm ascii armoring to binary data"""

		bin = ""
		for i in range(len(data) / 2):
			high = ord(data[2*i]) - ord("a")
			low = ord(data[2*i+1]) - ord("a")
			bin = bin + chr(high*16+low)

		return bin


	def __rotate_field(self, data):
		"""Rotates a field"""

		blocks = int(math.ceil(len(data) / float(self.blocksize)))

		scrambled = ""
		for block in range(blocks):
			for offset in range(self.blocksize):
				scrambled = scrambled + data[offset * blocks + block]

		return scrambled


	def __unrotate_field(self, data):
		"""Unrotates a field"""

		blocks = int(math.ceil(len(data) / float(self.blocksize)))

		plain = ""
		for offset in range(self.blocksize):
			for block in range(blocks):
				char = data[block * self.blocksize + offset]
				if char == "\x00":
					break

				plain = plain + char

			if char == "\x00":
				break

		return plain


	def __xml_export_convert_data(self, data):
		"""Converts data to fpm data structure"""

		fields = data["fields"]

		fpmdata = {
			"title"		: data["name"],
			"user"		: "",
			"url"		: "",
			"password"	: "",
			"notes"		: data["description"],
			"category"	: "",
			"launcher"	: ""
		}

		if data["type"] == revelation.data.ENTRY_ACCOUNT_CREDITCARD:
			fpmdata["user"]		= fields[revelation.data.FIELD_CREDITCARD_CARDNUMBER]
			fpmdata["password"]	= fields[revelation.data.FIELD_GENERIC_PIN]

		elif data["type"] == revelation.data.ENTRY_ACCOUNT_CRYPTOKEY:
			fpmdata["user"]		= fields[revelation.data.FIELD_GENERIC_KEYFILE]
			fpmdata["password"]	= fields[revelation.data.FIELD_GENERIC_PASSWORD]
			fpmdata["url"]		= fields[revelation.data.FIELD_GENERIC_HOSTNAME]

		elif data["type"] == revelation.data.ENTRY_ACCOUNT_DATABASE:
			fpmdata["user"]		= fields[revelation.data.FIELD_GENERIC_USERNAME]
			fpmdata["password"]	= fields[revelation.data.FIELD_GENERIC_PASSWORD]
			fpmdata["url"]		= fields[revelation.data.FIELD_GENERIC_HOSTNAME] + ": " + fields[revelation.data.FIELD_GENERIC_DATABASE]

		elif data["type"] == revelation.data.ENTRY_ACCOUNT_DOOR:
			fpmdata["password"]	= fields[revelation.data.FIELD_GENERIC_CODE]
			fpmdata["url"]		= fields[revelation.data.FIELD_GENERIC_LOCATION]

		elif data["type"] == revelation.data.ENTRY_ACCOUNT_EMAIL:
			fpmdata["user"]		= fields[revelation.data.FIELD_GENERIC_USERNAME]
			fpmdata["password"]	= fields[revelation.data.FIELD_GENERIC_PASSWORD]
			fpmdata["url"]		= fields[revelation.data.FIELD_GENERIC_HOSTNAME]

		elif data["type"] == revelation.data.ENTRY_ACCOUNT_FTP:
			fpmdata["user"]		= fields[revelation.data.FIELD_GENERIC_USERNAME]
			fpmdata["password"]	= fields[revelation.data.FIELD_GENERIC_PASSWORD]
			fpmdata["url"]		= "ftp://" + fields[revelation.data.FIELD_GENERIC_HOSTNAME]
	
			if fields[revelation.data.FIELD_GENERIC_PORT] != "":
				fpmdata["url"]	= fpmdata["url"] + ":" + fields[revelation.data.FIELD_GENERIC_PORT]

		elif data["type"] == revelation.data.ENTRY_ACCOUNT_GENERIC:
			fpmdata["user"]		= fields[revelation.data.FIELD_GENERIC_USERNAME]
			fpmdata["password"]	= fields[revelation.data.FIELD_GENERIC_PASSWORD]

		elif data["type"] == revelation.data.ENTRY_ACCOUNT_PHONE:
			fpmdata["user"]		= fields[revelation.data.FIELD_PHONE_PHONENUMBER]
			fpmdata["password"]	= fields[revelation.data.FIELD_GENERIC_PIN]

		elif data["type"] == revelation.data.ENTRY_ACCOUNT_SHELL:
			fpmdata["user"]		= fields[revelation.data.FIELD_GENERIC_USERNAME]
			fpmdata["password"]	= fields[revelation.data.FIELD_GENERIC_PASSWORD]
			fpmdata["url"]		= fields[revelation.data.FIELD_GENERIC_HOSTNAME]

		elif data["type"] == revelation.data.ENTRY_ACCOUNT_USENET:
			fpmdata["user"]		= fields[revelation.data.FIELD_GENERIC_USERNAME]
			fpmdata["password"]	= fields[revelation.data.FIELD_GENERIC_PASSWORD]
			fpmdata["url"]		= fields[revelation.data.FIELD_GENERIC_HOSTNAME]

		elif data["type"] == revelation.data.ENTRY_ACCOUNT_WEBSITE:
			fpmdata["user"]		= fields[revelation.data.FIELD_GENERIC_USERNAME]
			fpmdata["password"]	= fields[revelation.data.FIELD_GENERIC_PASSWORD]
			fpmdata["url"]		= fields[revelation.data.FIELD_GENERIC_URL]

		return fpmdata


	def __xml_export_entries(self, entrystore, parent = None):
		"""Exports xml for entries"""

		xml = ""

		for i in range(entrystore.iter_n_children(parent)):
			child = entrystore.iter_nth_child(parent, i)

			data = entrystore.get_entry(child)

			if data["type"] == revelation.data.ENTRY_FOLDER:
				xml = xml + self.__xml_export_entries(entrystore, child)

			else:

				# convert data to fpm structure
				fpmdata = self.__xml_export_convert_data(data)

				path = entrystore.get_path(child)
				if len(path) > 1:
					fpmdata["category"] = entrystore.get_entry(entrystore.get_iter(path[0]))["name"]

				xml = xml + "<PasswordItem>"
				xml = xml + "<title>" + self.__encrypt_field(fpmdata["title"]) + "</title>"
				xml = xml + "<user>" + self.__encrypt_field(fpmdata["user"]) + "</user>"
				xml = xml + "<url>" + self.__encrypt_field(fpmdata["url"]) + "</url>"
				xml = xml + "<password>" + self.__encrypt_field(fpmdata["password"]) + "</password>"
				xml = xml + "<notes>" + self.__encrypt_field(fpmdata["notes"]) + "</notes>"
				xml = xml + "<category>" + self.__encrypt_field(fpmdata["category"]) + "</category>"
				xml = xml + "<launcher>" + self.__encrypt_field(fpmdata["launcher"]) + "</launcher>"
				xml = xml + "</PasswordItem>"

		return xml


	def __xml_import_entry(self, entry, entrystore, foldercache):
		"""Imports an entry into the entrystore"""

		if entry.type != "element" or entry.name != "PasswordItem":
			return

		parent = None
		data = {
			"type"		: revelation.data.ENTRY_ACCOUNT_GENERIC,
			"updated"	: time.time(),
			"fields"	: {}
		}

		field = entry.children
		while field is not None:
			if field.type == "element":
				content = self.__decrypt_field(field.content)

				if field.name == "title":
					data["name"] = content

				elif field.name == "user":
					data["fields"][revelation.data.FIELD_GENERIC_USERNAME] = content

				elif field.name == "url":
					data["fields"][revelation.data.FIELD_GENERIC_HOSTNAME] = content

				elif field.name == "password":
					data["fields"][revelation.data.FIELD_GENERIC_PASSWORD] = content

				elif field.name == "notes":
					data["description"] = content

				elif field.name == "category":
					foldername = content.strip()

					if foldername == "":
						pass
					elif foldercache.has_key(foldername):
						parent = foldercache[foldername]
					else:
						parent = entrystore.add_entry(None, {"type": revelation.data.ENTRY_FOLDER, "name": foldername, "updated": time.time()})
						foldercache[foldername] = parent

			field = field.next

		entrystore.add_entry(parent, data)


	def __xml_import_keyinfo(self, root):
		"""Extracts keyinfo from the XML"""

		keynode = self.xml_import_scan(root, "KeyInfo")
		if keynode == None:
			raise FormatError

		attrs = self.xml_import_attrs(keynode)
		if not attrs.has_key("salt") or not attrs.has_key("vstring"):
			raise FormatError

		info = {
			"salt"		: attrs["salt"],
			"vstring"	: attrs["vstring"]
		}

		return info


	def check_data(self, data):
		"""Checks if the file header is valid"""

		if re.search("^<\?xml.*\?>\s*<FPM\s+", data, re.IGNORECASE) == None:
			raise FormatError


	def detect_type(self, header):
		"""Checks if the file is a FPM file, based on a header"""

		try:
			self.check_data(header)

		except FormatError:
			return gtk.FALSE

		else:
			return gtk.TRUE


	def export_data(self, entrystore, options):
		"""Exports data from an entrystore"""

		salt = self.__generate_salt()
		self.cipher_init(Blowfish, MD5.new(salt + options[OPTION_PASSWORD]).digest())

		data = "<?xml version=\"1.0\"?>\n"
		data = data + "<FPM full_version=\"00.58.00\" min_version=\"00.58.00\" display_version=\"0.58\">"
		data = data + "<KeyInfo salt=\"" + salt + "\" vstring=\"" + self.__encrypt_field("FIGARO") + "\"/>"
		data = data + "<LauncherList><LauncherItem></LauncherItem></LauncherList>"
		data = data + "<PasswordList>"
		data = data + self.__xml_export_entries(entrystore)
		data = data + "</PasswordList>"
		data = data + "</FPM>"

		return data


	def import_data(self, entrystore, data, options):
		"""Imports data into an entrystore"""

		doc = self.xml_import_init(data)

		# fetch root node, and check validity
		root = doc.children
		if root.name != "FPM":
			raise FormatError

		# set up cipher and check if password is correct
		keyinfo = self.__xml_import_keyinfo(root)
		self.cipher_init(Blowfish, MD5.new(keyinfo["salt"] + options[OPTION_PASSWORD]).digest())

		if self.__decrypt_field(keyinfo["vstring"]) != "FIGARO":
			raise PasswordError

		# import entries into entrystore
		pwlist = self.xml_import_scan(root, "PasswordList")
		if pwlist == None:
			raise FormatError

		foldercache = {}
		entry = pwlist.children
		while entry is not None:
			self.__xml_import_entry(entry, entrystore, foldercache)
			entry = entry.next



class RevelationXMLData(Data):
	"""Class for handling Revelation XML data"""

	def __xml_export(self, entrystore, parent = None, level = 0):
		"""Exports XML based on an entrystore"""

		xml = ""
		tabs = "\t" * (level + 1)

		# process each child
		for i in range(entrystore.iter_n_children(parent)):
			iter = entrystore.iter_nth_child(parent, i)
			data = entrystore.get_entry(iter)

			xml = xml + "\n"
			xml = xml + tabs + "<entry type=\"" + revelation.data.escape_markup(data["type"]) + "\">\n"
			xml = xml + tabs + "	<name>" + revelation.data.escape_markup(data["name"]) + "</name>\n"
			xml = xml + tabs + "	<description>" + revelation.data.escape_markup(data["description"]) + "</description>\n"
			xml = xml + tabs + "	<updated>" + revelation.data.escape_markup(str(data["updated"])) + "</updated>\n"

			for field, value in data["fields"].items():
				xml = xml + tabs + "	<field id=\"" + revelation.data.escape_markup(field) + "\">" + revelation.data.escape_markup(value) + "</field>\n"

			# handle any children
			if entrystore.iter_n_children(iter) > 0:
				xml = xml + self.__xml_export(entrystore, iter, level + 1)

			xml = xml + tabs + "</entry>\n"

		# generate header and footer if at level 0
		if level == 0:
			header = "<?xml version=\"1.0\" ?>\n"
			header = header + "<revelationdata version=\"" + revelation.VERSION + "\" dataversion=\"" + str(revelation.DATAVERSION) + "\">\n"
			footer = "</revelationdata>\n"
			xml = header + xml + footer

		return xml


	def __xml_import(self, xml, entrystore):
		"""Imports XML data into an entrystore"""

		doc = self.xml_import_init(xml)

		# fetch and validate root
		root = doc.children
		if root.name != "revelationdata":
			raise FormatError

		attrs = self.xml_import_attrs(root)
		if attrs.has_key("dataversion") and int(attrs["dataversion"]) > revelation.DATAVERSION:
			raise VersionError

		# process entries
		child = root.children
		while child is not None:
			self.__xml_import_node(entrystore, child)
			child = child.next


	def __xml_import_node(self, entrystore, node, parent = None):
		"""Recursively imports an xml node into an entry store"""

		# skip text outside nodes
		if node.type == "text":
			return

		# skip unknown nodes
		if node.type != "element" or node.name != "entry":
			raise FormatError

		# set up initial data structure
		data = {
			"type"		: self.xml_import_attrs(node)["type"],
			"fields"	: {}
		}

		if revelation.data.EntryTypes().type_exists(data["type"]) == 0:
			raise EntryTypeError

		# add empty entry, iter needed for any children
		iter = entrystore.add_entry(parent)

		# get elements
		child = node.children
		while child is not None:

			if child.type == "element":

				if child.name == "name":
					data["name"] = child.content

				elif child.name == "description":
					data["description"] = child.content

				elif child.name == "updated":
					data["updated"] = int(child.content)

				elif child.name == "field":
					field = self.xml_import_attrs(child)["id"]

					if entrystore.entrytypes.field_exists(data["type"], field) == 0:
						raise EntryFieldError

					data["fields"][field] = child.content

				elif child.name == "entry":
					self.__xml_import_node(entrystore, child, iter)

				else:
					raise FormatError

			child = child.next

		# update entry with actual data
		entrystore.update_entry(iter, data)


	def check_data(self, data):
		"""Checks for a valid data header"""

		match = re.search("^\s*<\?xml.*\?>\s*<revelationdata.+dataversion=\"(\S+)\"", data, re.IGNORECASE)

		if match == None:
			raise FormatError

		if int(match.group(1)) > revelation.DATAVERSION:
			raise VersionError


	def detect_type(self, data):
		"""Checks if the data is Revelation data"""

		try:
			self.check_data(data)

		except FormatError:
			return gtk.FALSE

		else:
			return gtk.TRUE


	def export_data(self, entrystore, options):
		"""Exports data from an entrystore"""

		return self.__xml_export(entrystore)


	def import_data(self, entrystore, data, options):
		"""Imports data into an entrystore"""

		self.__xml_import(data, entrystore)



class RevelationData(RevelationXMLData):
	"""Class for handling Revelation data files"""

	def __decrypt(self, data, password, iv = None):
		"""Decrypts data"""

		self.cipher_init(AES, password, iv, 16, 32)
		return self.cipher_decrypt(data)


	def __encrypt(self, data, password, iv = None):
		"""Encrypts data"""

		self.cipher_init(AES, password, iv, 16, 32)
		return self.cipher_encrypt(data)


	def __generate_header(self):
		"""Generates a new header for the data"""

		header = "rvl\x00" + chr(revelation.DATAVERSION) + "\x00"
		for part in revelation.VERSION.split("."):
			header = header + chr(int(part))
		header = header + "\x00\x00\x00"

		return header


	def __parse_header(self, header):
		"""Parses the data header, returns a tuple with the data version and app version"""

		if len(header) != 12 or header[0:3] != "rvl":
			return None

		dataversion = ord(header[4])
		appversion = str(ord(header[6])) + "." + str(ord(header[7])) + "." + str(ord(header[8]))

		return dataversion, appversion


	def check_data(self, data):
		"""Checks if the file header is valid"""

		headerdata = self.__parse_header(data[0:12])

		# ignore version 0 data files (deprecated, remove in future version)
		if headerdata == None:
			return

		if headerdata[0] > revelation.DATAVERSION:
			raise VersionError


	def detect_type(self, header):
		"""Checks if the header is a valid Revelation data file"""

		if self.__parse_header(header[0:12]) == None:
			return gtk.FALSE
		else:
			return gtk.TRUE


	def export_data(self, entrystore, options):
		"""Exports data from an entrystore"""

		# first, generate XML data from the entrystore
		data = RevelationXMLData.export_data(self, entrystore, options[OPTION_PASSWORD])

		# next, compress the data and right-pad it
		# (the pad is the repeated ascii value of the pad length)
		data = zlib.compress(data)

		padlen = 16 - (len(data) % 16)
		if padlen == 0:
			padlen = 16

		data = data + (chr(padlen) * padlen)

		# generate an initial vector for CBC encryption mode
		iv = ""
		for i in range(16):
			iv = iv + chr(int(random.random() * 255))

		# encrypt the data
		data = self.__encrypt(data, options[OPTION_PASSWORD], iv)

		# encrypt the iv, and prepend it to the data along with a header
		data = self.__generate_header() + self.__encrypt(iv, options[OPTION_PASSWORD]) + data

		return data


	def import_data(self, entrystore, data, options):
		"""Imports the data into an entrystore"""

		# get the version numbers from the header
		headerdata = self.__parse_header(data[0:12])

		# if no valid header was found, assume version 0 (0.1.x series).
		# this is deprecated, and support will be removed in a future version
		if headerdata == None:

			data = self.__decrypt(data, options[OPTION_PASSWORD])

			# if not xml file, assume wrong password (could be invalid file
			# as well, but in most cases it will be a wrong password)
			if data[0:5] != "<?xml":
				raise PasswordError

			RevelationXMLData.import_data(self, entrystore, data, options)


		# handle version 1 data file
		elif headerdata[0] == 1:

			# fetch and decrypt the initial vector for CBC mode decryption
			iv = self.__decrypt(data[12:28], options[OPTION_PASSWORD])

			# decrypt the data
			data = self.__decrypt(data[28:], options[OPTION_PASSWORD], iv)

			# get and check the pad length
			padlen = ord(data[-1])
			for i in data[-padlen:]:
				if ord(i) != padlen:
					raise PasswordError

			# decompress data
			data = zlib.decompress(data[0:-padlen])

			if data[0:5] != "<?xml":
				raise PasswordError

			# import xml into entrystore
			RevelationXMLData.import_data(self, entrystore, data, options)


		# future file version, raise exception
		else:
			raise VersionError

