#
# Revelation 0.2.1 - a password manager for GNOME 2
# http://oss.wired-networks.net/revelation/
#
# Module containing dialog classes
#
#
# Copyright (c) 2003-2004 Erik Grinaker
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#

import gtk, gnome, gnome.ui, revelation, time

# first we define a few base classes

class Dialog(gtk.Dialog):
	"""A base dialog class"""

	def __init__(self, parent, title, buttons, default = None):

		gtk.Dialog.__init__(self, title, parent, gtk.DIALOG_DESTROY_WITH_PARENT | gtk.DIALOG_MODAL)

		self.set_border_width(6)
		self.vbox.set_spacing(12)
		self.set_has_separator(gtk.FALSE)
		self.set_resizable(gtk.FALSE)

		self.__init_buttons(buttons, default)


	def __init_buttons(self, buttons, default):
		"""Sets up buttons for the dialog"""

		for button, i in zip(buttons, range(len(buttons))):
			b = self.add_button(button[0], button[1])

			# set default button
			if default == i or (default == None and i == 0):
				self.set_default_response(button[1])


	def get_button(self, index):
		"""Returns a button"""

		buttons = self.get_children()[0].get_children()[-1].get_children()

		if index < len(buttons):
			return buttons[index]
		else:
			return None



class FileSelector(gtk.FileSelection):

	def __init__(self, title = None):
		gtk.FileSelection.__init__(self, title)


	def run(self):
		"""displays the file selector"""

		self.show()

		if gtk.FileSelection.run(self) == gtk.RESPONSE_OK:
			filename = self.get_filename()
			self.destroy()
			return filename
		else:
			self.destroy()
			raise revelation.CancelError



class Hig(Dialog):
	"""HIGified dialog base class"""

	def __init__(self, parent, pritext, sectext, stockimage, buttons, default = None):

		Dialog.__init__(self, parent, "", buttons, default)
		self.__init_contents(pritext, sectext, stockimage)


	def __init_contents(self, pritext, sectext, stockimage):
		"""Sets up the content area of the dialog"""

		# hbox separating dialog image and contents
		hbox = gtk.HBox()
		hbox.set_spacing(12)
		hbox.set_border_width(6)
		self.vbox.pack_start(hbox)

		# set up image
		if stockimage != None:
			image = gtk.Image()
			image.set_from_stock(stockimage, gtk.ICON_SIZE_DIALOG)
			image.set_alignment(0.5, 0)
			hbox.pack_start(image)

		# set up main content area
		self.contents = gtk.VBox()
		self.contents.set_spacing(10)
		hbox.pack_start(self.contents)

		# set contents
		label = gtk.Label("<span size=\"larger\" weight=\"bold\">" + revelation.data.escape_markup(pritext) + "</span>\n\n" + sectext)
		label.set_use_markup(gtk.TRUE)
		label.set_line_wrap(gtk.TRUE)
		label.set_alignment(0, 0)
		self.contents.pack_start(label)


	def run(self):
		"""Runs the dialog, and returns the response"""

		self.show_all()
		response = gtk.Dialog.run(self)
		self.destroy()

		return response


class Property(Dialog):
	"""Base class for property dialogs, such as item properties or preferences"""

	def __init__(self, parent, title, buttons, default = None):

		Dialog.__init__(self, parent, title, buttons, default)

		self.set_border_width(12)
		self.vbox.set_spacing(18)

		self.sizegroup = gtk.SizeGroup(gtk.SIZE_GROUP_HORIZONTAL)


	def add_section(self, title, description = None):
		"""Adds an inputsection to the dialog"""

		section = revelation.ui.InputSection(title, self.sizegroup, description)
		self.vbox.pack_start(section)

		return section


# the following classes may be subclassed from the bases above

class About(gnome.ui.About):
	"""Displays an about dialog"""

	def __init__(self):

		gnome.ui.About.__init__(
			self, revelation.APPNAME, revelation.VERSION, revelation.COPYRIGHT,
			"\"" + revelation.RELNAME + "\"\n\nRevelation is a password manager for the GNOME 2 desktop.",
			[ revelation.AUTHOR ], None, "",
			gtk.gdk.pixbuf_new_from_file(revelation.DATADIR + "/pixmaps/revelation.png")
		)


	def run(self):
		self.show_all()


class EditEntry(Property):
	"""Dialog for editing account entries"""

	def __init__(self, parent, title, entrytypes, data = None):

		Property.__init__(self, parent, title, [ [ gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL ], [ gtk.STOCK_OK, gtk.RESPONSE_OK ] ], 1)

		# set a few properties
		self.entrytypes = entrytypes
		self.tooltips = gtk.Tooltips()
		self.fieldsection = None
		self.entry_field = {}
		self.fielddata = {}
		self.curtype = None

		# set up meta section
		self.__init_metasection(title)

		# set initial states
		if data == None:
			self.update(revelation.data.ENTRY_FOLDER)
		else:
			self.entry_name.set_text(data["name"])
			self.entry_desc.set_text(data["description"])
			self.update(data["type"], data["fields"])


	def __init_metasection(self, title):
		"""Sets up the meta section"""

		self.metasection = self.add_section(title)

		self.entry_name = self.__ui_textentry(None, "The name of the entry")
		self.metasection.add_inputrow("Name", self.entry_name)

		self.entry_desc = self.__ui_textentry(None, "A description of the entry")
		self.metasection.add_inputrow("Description", self.entry_desc)

		self.dropdown = self.__ui_dropdown()
		self.dropdown.connect("changed", self.__cb_update)
		self.metasection.add_inputrow("Type", self.dropdown)


	def __ui_dropdown(self):
		"""Creates the dropdown menu for entry types"""

		typelist = self.entrytypes.get_list()
		typelist.remove(revelation.data.ENTRY_FOLDER)
		typelist.insert(0, revelation.data.ENTRY_FOLDER)

		dropdown = revelation.ui.DropDown()
		self.tooltips.set_tip(dropdown, "The type of account - folders can contain other entries")

		for type in typelist:
			item = dropdown.append_imageitem(self.entrytypes.get_data(type)["name"], self.entrytypes.get_data(type)["icon"])
			item.type = type

			if type == revelation.data.ENTRY_FOLDER:
				dropdown.append_separator()

		return dropdown


	def __ui_textentry(self, value = None, tooltip = None):
		"""Creates a entry widget for normal text input"""

		entry = gtk.Entry()
		entry.set_width_chars(50)
		entry.set_activates_default(gtk.TRUE)

		if value != None:
			entry.set_text(value)

		if tooltip != None:
			self.tooltips.set_tip(entry, tooltip)

		return entry


	def __cb_update(self, object):
		"""Callback for updating the field form"""

		# only update field form if type has changed
		type = self.get_type()
		if self.curtype != type:
			self.update(type)
			self.show_all()


	def get_type(self):
		"""Fetches the currently selected type from the dropdown"""
		return self.dropdown.get_active_item().type


	def get_data(self):
		"""Returns the current data from the dialog"""

		data = {
			"name"		: self.entry_name.get_text(),
			"description"	: self.entry_desc.get_text(),
			"type"		: self.get_type(),
			"icon"		: self.entrytypes.get_data(self.get_type())["icon"],
			"updated"	: time.time(),
			"fields"	: {}
		}

		for field in self.entrytypes.get_data(data["type"])["fields"]:
			if self.entry_field.has_key(field):
				value = self.entry_field[field].get_text()
			else:
				value = ""

			data["fields"][field] = value

		return data


	def run(self):
		"""Displays the dialog, and returns the data"""

		self.show_all()

		# run dialog and fetch data
		if Property.run(self) == gtk.RESPONSE_OK:
			data = self.get_data()

			# check if name is filled in
			if data["name"] == "":
				Error(self, "No name given", "You need to enter a name for the entry.").run()
				return self.run()

			self.destroy()
			return data

		else:
			self.destroy()
			raise revelation.CancelError


	def set_type(self, type):
		"""Sets the current type for the dialog"""

		for item in self.dropdown.get_menu().get_children():
			if hasattr(item, "type") and item.type == type:
				self.dropdown.set_active_item(item)

		# store current type
		self.curtype = type


	def set_typechange_allowed(self, allow):
		"""Sets whether it is allowed to change the type"""

		# get menu items, without separator
		items = self.dropdown.get_menu().get_children()

		for item in items:
			if allow == gtk.FALSE and hasattr(item, "type") and item.type != self.curtype:
				item.set_sensitive(gtk.FALSE)
			else:
				item.set_sensitive(gtk.TRUE)


	def update(self, type, fielddata = None):
		"""Updates the field form for the selected type"""

		# update accumulated field data with current data
		self.fielddata.update(self.get_data()["fields"])

		if fielddata != None:
			for key, value in fielddata.items():

				# dont update with empty values
				if value != "":
					self.fielddata[key] = value

		self.set_type(type)
		fields = self.entrytypes.get_data(type)["fields"]

		# don't set up field section if not needed
		if len(fields) == 0:

			if self.fieldsection != None:
				self.fieldsection.destroy()
				self.fieldsection = None

		else:

			if self.fieldsection == None:
				self.fieldsection = self.add_section("Account data")

			self.fieldsection.clear()

			# set up fields
			for field in fields:

				if fielddata != None and fielddata.has_key(field) and fielddata[field] != "":
					value = fielddata[field]
				elif self.fielddata.has_key(field):
					value = self.fielddata[field]
				else:
					value = ""

				self.entry_field[field] = self.__ui_textentry(value, self.entrytypes.get_field_data(field)["tooltip"])
				self.fieldsection.add_inputrow(self.entrytypes.get_field_data(field)["name"], self.entry_field[field])


class Error(Hig):

	def __init__(self, parent, pritext, sectext):
		Hig.__init__(
			self, parent, pritext, sectext, gtk.STOCK_DIALOG_ERROR,
			[ [ gtk.STOCK_OK, gtk.RESPONSE_OK ] ]
		)



class FileOverwrite(Hig):
	"""Asks for confirmation when overwriting existing file"""

	def __init__(self, parent, file):
		Hig.__init__(
			self, parent, "Overwrite existing file?",
			"The file '" + file + "' already exists. If you choose to overwrite the file, its contents will be lost.",
			gtk.STOCK_DIALOG_WARNING, [ [ gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL ], [ revelation.ui.STOCK_OVERWRITE, gtk.RESPONSE_OK ] ]
		)


	def run(self):
		"""Runs the dialog"""

		return Hig.run(self) == gtk.RESPONSE_OK



class Find(Dialog):
	"""Dialog for searching for entries"""

	def __init__(self, parent):
		Dialog.__init__(
			self, parent, "Find an entry",
			[ [ gtk.STOCK_CLOSE, gtk.RESPONSE_CLOSE ], [ revelation.ui.STOCK_PREVIOUS, revelation.ui.RESPONSE_PREVIOUS ], [ revelation.ui.STOCK_NEXT, revelation.ui.RESPONSE_NEXT ] ], 2
		)

		section = revelation.ui.InputSection("Find an entry:")
		self.vbox.pack_start(section)

		self.entry_phrase = gtk.Entry()
		self.entry_phrase.set_activates_default(gtk.TRUE)
		self.entry_phrase.connect("changed", self.__cb_entry_changed)
		section.add_inputrow("Search for", self.entry_phrase)

		self.dropdown = self.__init_dropdown()
		self.dropdown.connect("changed", self.__cb_dropdown_changed)
		section.add_inputrow("Account type", self.dropdown)

		self.check_folders = gtk.CheckButton("Search for folders as well")
		self.check_folders.set_active(gtk.TRUE)
		section.add_inputrow(None, self.check_folders)

		self.check_namedesc = gtk.CheckButton("Only search in name and description")
		section.add_inputrow(None, self.check_namedesc)

		self.check_casesens = gtk.CheckButton("Case sensitive")
		section.add_inputrow(None, self.check_casesens)


		# set up initial states
		self.entry_phrase.emit("changed")


	def __init_dropdown(self):
		dropdown = revelation.ui.DropDown()

		item = dropdown.append_imageitem("Any", "gnome-stock-about")
		item.type = None

		dropdown.append_separator()

		entrytypes = revelation.data.EntryTypes()
		for type in entrytypes.get_list():

			if type == revelation.data.ENTRY_FOLDER:
				continue

			typedata = entrytypes.get_data(type)
			item = dropdown.append_imageitem(typedata["name"], typedata["icon"])
			item.type = type

		return dropdown


	def __cb_dropdown_changed(self, widget):

		if self.get_type() == None:
			self.check_folders.set_sensitive(gtk.TRUE)
		else:
			self.check_folders.set_active(gtk.FALSE)
			self.check_folders.set_sensitive(gtk.FALSE)


	def __cb_entry_changed(self, widget):

		if len(self.entry_phrase.get_text()) == 0:
			self.get_button(0).set_sensitive(gtk.FALSE)
			self.get_button(1).set_sensitive(gtk.FALSE)
		else:
			self.get_button(0).set_sensitive(gtk.TRUE)
			self.get_button(1).set_sensitive(gtk.TRUE)


	def get_type(self):
		return self.dropdown.get_active_item().type


	def set_type(self, type):
		for child in self.dropdown.get_menu().get_children():
			if hasattr(child, "type") and child.type == type:
				self.dropdown.set_active_item(child)


	def run(self):
		"""runs the dialog"""

		self.show_all()
		return Dialog.run(self)


class Password(Hig):
	"""Displays a dialog asking for a password"""

	def __init__(self, parent, title, text, confirm = gtk.FALSE, current = None):

		Hig.__init__(
			self, parent, title, text, revelation.ui.STOCK_PASSWORD,
			[ [ gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL ], [ gtk.STOCK_OK, gtk.RESPONSE_OK ] ], 1
		)

		self.confirm = confirm
		self.current = current

		inputsection = revelation.ui.InputSection()
		self.contents.pack_start(inputsection)

		# ask for current password
		if current != None:
			self.entry_current = self.ui_passwordentry()
			inputsection.add_inputrow("Current", self.entry_current)

		# password entry
		if current != None: # assume new password if asking for current
			label = "New password"
		else:
			label = "Password"
		self.entry_password = self.ui_passwordentry()
		inputsection.add_inputrow(label, self.entry_password)

		# ask for password confirmation
		if confirm == gtk.TRUE:
			self.entry_confirm = self.ui_passwordentry()
			inputsection.add_inputrow("Confirm", self.entry_confirm)


	def ui_passwordentry(self):
		"""Creates a password entry field"""

		entry = gtk.Entry()
		entry.set_visibility(gtk.FALSE)
		entry.set_activates_default(gtk.TRUE)

		return entry


	def run(self):

		self.show_all()

		# get password
		if Property.run(self) == gtk.RESPONSE_OK:
			password = self.entry_password.get_text()

			# check if current password is correct
			if self.current != None and self.current != self.entry_current.get_text():
				Error(self, "Incorrect password", "The password you entered as the current file password is incorrect.").run()
				return self.run()

			# display error if no password
			if len(password) == 0:
				Error(self, "Password not given", "You need to enter a password.").run()
				return self.run()

			# check if confirm password matches
			if self.confirm == gtk.TRUE and password != self.entry_confirm.get_text():
				Error(self, "Passwords don't match", "The password and password confirmation you entered does not match.").run()
				return self.run()

			self.destroy()
			return password

		else:
			self.destroy()
			raise revelation.CancelError



class RemoveEntry(Hig):
	"""Dialog for confirming entry removal"""

	def __init__(self, parent, data):

		if isinstance(data, dict):

			if data["type"] == revelation.data.ENTRY_FOLDER:
				pritext = "Really remove folder '" + data["name"] + "'?"
				sectext = "By removing this folder you will also remove all accounts and folders it contains."
			else:
				pritext = "Really remove account '" + data["name"] + "'?"
				sectext = "Please confirm that you wish to remove this account."

		elif isinstance(data, int):
			pritext = "Really remove the " + str(data) + " selected entries?"
			sectext = "By removing these entries you will also remove any entries they may contain."

		Hig.__init__(
			self, parent, pritext, sectext, gtk.STOCK_DIALOG_WARNING,
			[ [ gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL ], [ revelation.ui.STOCK_REMOVE, gtk.RESPONSE_OK ] ]
		)


	def run(self):
		"""Runs the dialog, returns true if entries are to be removed"""

		return Hig.run(self) == gtk.RESPONSE_OK

