#
# Revelation 0.2.1 - a password manager for GNOME 2
# http://oss.wired-networks.net/revelation/
#
# Module containing druid classes
#
#
# Copyright (c) 2003-2004 Erik Grinaker
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#

import gtk, gnome.ui, revelation, os

class Page(gnome.ui.DruidPageStandard):
	"""Class for normal druid pages"""

	def __init__(self, title, logo = None):
		gnome.ui.DruidPageStandard.__init__(self)

		self.set_title(title)

		if logo == None:
			logo = self.render_icon(revelation.ui.STOCK_APPLICATION, revelation.ui.ICON_SIZE_DRUID)

		self.set_logo(logo)


	def append_item(self, text, widget, extratext = None):
		"""Adds an item to the druid page"""

		vbox = gtk.VBox()
		vbox.set_spacing(18)
		vbox.set_border_width(6)
		gnome.ui.DruidPageStandard.append_item(self, "", vbox, "")

		if text != None:
			label = gtk.Label(text)
			label.set_line_wrap(gtk.TRUE)
			label.set_justify(gtk.JUSTIFY_CENTER)
			vbox.pack_start(label)

		vbox.pack_start(widget, gtk.FALSE, gtk.FALSE)

		if extratext != None:
			label = gtk.Label()
			label.set_markup("<span size=\"small\">" + extratext + "</span>")
			label.set_line_wrap(gtk.TRUE)
			label.set_justify(gtk.JUSTIFY_CENTER)
			vbox.pack_start(label)



class PageEdge(gnome.ui.DruidPageEdge):
	"""Class for druid edge pages"""

	def __init__(self, position, title, text = None, logo = None):
		gnome.ui.DruidPageEdge.__init__(self, position)

		self.set_title(title)
		self.pagepos = position

		if text != None:
			self.set_text(text)

		if logo == None:
			logo = self.render_icon(revelation.ui.STOCK_APPLICATION, revelation.ui.ICON_SIZE_DRUID, None)

		self.set_logo(logo)



class Druid(gnome.ui.Druid):
	"""Base class for druids"""

	def __init__(self, parent, title):
		gnome.ui.Druid.__init__(self)
		self.connect("cancel", self.__cb_cancel)

		# set up a dialog for the druid
		self.dialog = gtk.Dialog(title, parent, (gtk.DIALOG_DESTROY_WITH_PARENT | gtk.DIALOG_MODAL | gtk.DIALOG_NO_SEPARATOR))
		self.dialog.action_area.destroy()
		self.dialog.vbox.pack_start(self)


	# callbacks
	def __cb_cancel(self, object):
		self.dialog.response(gtk.RESPONSE_CANCEL)

	def __cb_finish(self, object, data):
		self.dialog.response(gtk.RESPONSE_OK)

	def __cb_destroy_page(self, page, data):
		page.destroy()


	# main methods
	def append_page(self, page):
		"""Appends a page to the druid"""

		if hasattr(page, "pagepos") and page.pagepos == gnome.ui.EDGE_FINISH:
			page.connect("finish", self.__cb_finish)

		gnome.ui.Druid.append_page(self, page)


	def insert_page(self, prevpage, page):
		"""Inserts a page into the druid"""

		gnome.ui.Druid.insert_page(self, prevpage, page)
		page.connect("back", self.__cb_destroy_page)
		page.show_all()


	def run(self):
		"""Runs the druid"""

		self.dialog.show_all()
		response = self.dialog.run()
		self.destroy()
		self.dialog.destroy()

		return response



class ExportFile(Druid):
	"""Druid for exporting data"""

	def __init__(self, parent):
		Druid.__init__(self, parent, "Export File")

		self.filetypes = revelation.datafile.FileTypes()
		self.datafile = revelation.datafile.DataFile()

		self.append_page(self.__page_start())
		self.append_page(self.__page_file())
		self.append_page(self.__page_finish())


	# setup pages
	def __page_file(self):
		page = Page("Select File to Export to")
		page.connect("next", self.__cb_page_file)

		# set up input fields
		section = revelation.ui.InputSection()
		page.append_item("Select the file you wish to export data to, and the file type:", section)

		page.fileentry = gnome.ui.FileEntry("revelation-file-export", "Select File to Export to")
		page.fileentry.set_modal(gtk.TRUE)
		section.add_inputrow("File", page.fileentry)

		page.dropdown = revelation.ui.DropDown()
		section.add_inputrow("Filetype", page.dropdown)

		for type in self.filetypes.get_export_types():
			item = page.dropdown.append_item(self.filetypes.get_data(type, "name"))
			item.filetype = type

		return page


	def __page_start(self):
		return PageEdge(gnome.ui.EDGE_START, "File Export Assistant", "Welcome to the Revelation file export assistant, which will guide you through exporting accounts into a foreign file format.")


	def __page_finish(self):
		return PageEdge(gnome.ui.EDGE_FINISH, "Export File", "You have now provided all information needed to export the data. Press \"Apply\" to start the export.")


	def __page_password(self):

		page = Page("Enter Password")
		page.connect("next", self.__cb_page_password)

		# set up input fields
		section = revelation.ui.InputSection()
		page.append_item("Please enter a password to encrypt the file with:", section)

		page.passwordentry = gtk.Entry()
		page.confirmentry = gtk.Entry()

		for entry, name in zip((page.passwordentry, page.confirmentry), ("Password", "Confirm password")):
			entry.set_visibility(gtk.FALSE)
			entry.set_activates_default(gtk.TRUE)

			password = self.datafile.get_option(revelation.datafile.OPTION_PASSWORD)
			entry.set_text(password)

			section.add_inputrow(name, entry)

		return page



	# callbacks
	def __cb_page_file(self, page, data):
		"""Callback for checking file page input"""

		self.datafile.set_type(page.dropdown.get_active_item().filetype)
		file = page.fileentry.get_full_path(gtk.FALSE)

		if file == None:
			revelation.dialog.Error(self.dialog, "Filename not entered", "You need to enter a file to export the data to.").run()
			return gtk.TRUE

		if os.access(file, os.F_OK) == 1 and revelation.dialog.FileOverwrite(self.dialog, file).run() == gtk.FALSE:
			return gtk.TRUE

		self.datafile.set_file(file)

		# check if file format is insecure
		if self.filetypes.get_data(self.datafile.get_type(), "encryption") == gtk.FALSE:
			if revelation.dialog.Hig(
				self.dialog, "Export to insecure file?", "The file format you have chosen is not encrypted. If anyone has access to the file, they will be able to read your passwords.", gtk.STOCK_DIALOG_WARNING,
				[ [ gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL ], [ revelation.ui.STOCK_EXPORT, gtk.RESPONSE_OK ] ]
			).run() == gtk.RESPONSE_CANCEL:
				return gtk.TRUE

		# set up password page if a password is needed
		if self.datafile.has_option(revelation.datafile.OPTION_PASSWORD):
			self.insert_page(page, self.__page_password())


	def __cb_page_password(self, page, data):
		"""callback for checking the password page input"""

		# fetch passwords
		password = page.passwordentry.get_text()
		confirm = page.confirmentry.get_text()

		if len(password) == 0:
			revelation.dialog.Error(self.dialog, "No password given", "You must enter a password to decrypt the file.").run()
			return gtk.TRUE

		if password != confirm:
			revelation.dialog.Error(self.dialog, "Passwords don't match", "The passwords you entered do not match each other. Make sure that you enter the same password in both input fields.").run()
			return gtk.TRUE

		self.datafile.set_option(revelation.datafile.OPTION_PASSWORD, password)



	# main methods
	def run(self):
		"""runs the druid"""

		if Druid.run(self) == gtk.RESPONSE_OK:
			return self.datafile
		else:
			raise revelation.CancelError



class ImportFile(Druid):
	"""Druid for importing data"""

	def __init__(self, parent):
		Druid.__init__(self, parent, "Import File")

		self.filetypes = revelation.datafile.FileTypes()
		self.datafile = revelation.datafile.DataFile()

		self.append_page(self.__page_start())
		self.append_page(self.__page_file())
		self.append_page(self.__page_finish())


	# set up a few pages
	def __page_file(self):

		page = Page("Select File to Import")
		page.connect("next", self.__cb_page_file)

		# set up input fields
		section = revelation.ui.InputSection()
		page.append_item("Select the file you wish to import, and the type of file:", section, "You can select \"Autodetect\" to make Revelation attempt to find the filetype itself.")

		page.fileentry = gnome.ui.FileEntry("revelation-file-import", "Select File to Import")
		page.fileentry.set_modal(gtk.TRUE)
		section.add_inputrow("File", page.fileentry)

		page.dropdown = revelation.ui.DropDown()
		section.add_inputrow("Filetype", page.dropdown)

		item = page.dropdown.append_item("Autodetect")
		item.filetype = revelation.datafile.TYPE_AUTO
		page.dropdown.append_separator()

		for type in self.filetypes.get_import_types():
			item = page.dropdown.append_item(self.filetypes.get_data(type, "name"))
			item.filetype = type

		return page


	def __page_finish(self):
		return PageEdge(gnome.ui.EDGE_FINISH, "Import File", "You have now provided all required information to import the data. Press \"Apply\" to start the import.")


	def __page_password(self):

		page = Page("Enter Password")
		page.connect("next", self.__cb_page_password)

		# set up password field
		section = revelation.ui.InputSection()
		page.append_item("This file is encrypted - please enter a password to decrypt it:", section)

		page.passwordentry = gtk.Entry()
		page.passwordentry.set_visibility(gtk.FALSE)
		page.passwordentry.set_activates_default(gtk.TRUE)
		section.add_inputrow("Password", page.passwordentry)

		password = self.datafile.get_option(revelation.datafile.OPTION_PASSWORD)
		if password != "":
			page.passwordentry.set_text(password)

		return page


	def __page_start(self):
		return PageEdge(gnome.ui.EDGE_START, "File Import Assistant", "Welcome to the Revelation file import assistant, which will guide you through importing accounts from other password managers.")


	# callbacks
	def __cb_page_file(self, page, widget):
		"""Callback for checking the file page input"""

		# check the input file
		try:
			self.datafile.set_file(page.fileentry.get_full_path(gtk.FALSE))
			self.datafile.check_file()

			type = page.dropdown.get_active_item().filetype

			if type == revelation.datafile.TYPE_AUTO:
				self.datafile.detect_type()
			else:
				self.datafile.set_type(type)

			self.datafile.check_format()

		except IOError:
			revelation.dialog.Error(self.dialog, "Unable to open file", "The file you selected could not be opened. Please make sure the file exists, and that you have the proper permissions to open it.").run()
			return gtk.TRUE

		except revelation.datafile.FormatError:
			revelation.dialog.Error(self.dialog, "Invalid file format", "The file '" + self.datafile.get_file() + "' is not a valid " + self.filetypes.get_data(self.datafile.get_type(), "name") + " file.").run()
			return gtk.TRUE

		except revelation.datafile.VersionError:
			revelation.dialog.Error(self.dialog, "Future file version", "The file '" + self.datafile.get_file() + "' is from an unknown version. Please upgrade Revelation to a newer version to import it.").run()
			return gtk.TRUE

		except revelation.datafile.DetectError:
			revelation.dialog.Error(self.dialog, "Autodetection failed", "Revelation was not able to autodetect the format of the file '" + self.datafile.get_file() + "'. It may still be able to load the file, try specifying the file type manually.").run()
			return gtk.TRUE

		else:

			# set up password page if needed
			if self.datafile.has_option(revelation.datafile.OPTION_PASSWORD):
				self.insert_page(page, self.__page_password())


	def __cb_page_password(self, page, widget):
		"""Callback for checking the password page input"""

		password = page.passwordentry.get_text()

		if len(password) > 0:
			self.datafile.set_option(revelation.datafile.OPTION_PASSWORD, password)
			return gtk.FALSE
		else:
			revelation.dialog.Error(self.dialog, "No password given", "You must enter a password to decrypt the file.").run()
			return gtk.TRUE


	# main methods
	def run(self):
		"""Runs the druid"""

		if Druid.run(self) == gtk.RESPONSE_OK:
			return self.datafile
		else:
			raise revelation.CancelError

